import os
import requests
from logkit import log
import yaml


class Pax:
    
    PAX_ROOT           = ".pax/"
    PAX_CONFIG_DIR     = "~/.config/pax"
    PAX_NODE_FILE      = "pax_node.yml"
    PAX_ACTIONS_FILE   = "pax_actions.yml"
    GLOBAL_CONFIG_FILE = "pax_config.yml"

    INSTANCE = None

    @classmethod
    def instance(cls):
        if cls.INSTANCE is None:
            cls.INSTANCE = Pax()
        return cls.INSTANCE

    def __init__(self):
        self.api_endpoint = "http://google.com"
        log.info("Pax Singleton Initialized")

        try:
            data = self.__read_global_config()
            self.api_endpoint = data["api_endpoint"]
            log.info("Global Config Data Loaded", data)

        except Exception as e:
            log.info("Global Config not found. Please pax configure.", 
            {
                "error": str(e)
            })

    # ==========================================================================================
    # Public Methods
    # ==========================================================================================

    @staticmethod
    def get_remote_git_branch():
        process = os.popen('git config --get remote.origin.url')
        remote_origin = process.read()
        process.close()
        remote_url = remote_origin.replace(".git", "").strip().replace("\n", "")
        return remote_url

    # ==========================================================================================
    # File IO.
    # ==========================================================================================

    def ensure_and_join(self, directory: str, path: str):
        """ Expand the directory, ensure it exists, and join the path. """
        expanded_dir = os.path.expanduser(directory)
        os.makedirs(expanded_dir, exist_ok=True)
        full_path = os.path.join(expanded_dir, path)
        return full_path

    def load_local_object(self, path):
        object_file = self.ensure_and_join(self.PAX_ROOT, path)
        object_data = None
        if os.path.exists(object_file):
            with open(object_file, "r") as f:
                object_data = yaml.load(f)
        return object_data

    def save_local_object(self, path, data):
        object_file = self.ensure_and_join(self.PAX_ROOT, path)
        with open(object_file, "w") as f:
            yaml.dump(data, f)

    def load_pax_node_data(self):
        return self.load_local_object(self.PAX_NODE_FILE)

    def load_pax_actions_data(self):
        return self.load_local_object(self.PAX_ACTIONS_FILE)
    
    @property
    def node_data_dst(self):
        return self.ensure_and_join(self.PAX_ROOT, self.PAX_NODE_FILE)

    @property
    def actions_data_dst(self):
        return self.ensure_and_join(self.PAX_ROOT, self.PAX_ACTIONS_FILE)

    @property
    def global_config_dst(self):
        return self.ensure_and_join(self.PAX_CONFIG_DIR, self.GLOBAL_CONFIG_FILE)
  
    # ==========================================================================================
    # Server Comm. Methods.
    # ==========================================================================================

    def test_connection(self):
        status, response, data = self.api_get("")
        log.info("Response", {
            "status": status,
            "content": response.text
        })
        
        # Was the connection a success?
        # TODO: Add some other verification. Status might not be enough.
        return status == 200

    def get_tags(self):
        """ Get a list of package codes from the server. """
        _, _, data = self.api_get("get_nodes")
        nodes = data["nodes"]
        tags = {node["tag"] for node in nodes}
        log.info("Tags", tags)
        return tags

    def add_node_to_server(self, node):
        self.api_post("add_node", node)

    def api_get(self, route=""):
        endpoint = f"{self.api_endpoint}/{route}"
        print(f"Get Front Route: {endpoint}")
        response = requests.get(endpoint)
        return self.__spread(response)

    def api_post(self, route="", data={}):
        endpoint = f"{self.api_endpoint}/{route}"
        response = requests.post(endpoint, json=data)
        return self.__spread(response)

    def __spread(self, response):
        status = response.status_code
        try:
            data = response.json()
        except Exception as _:
            data = None
        return status, response, data

    # ==========================================================================================
    # Support Methods.
    # ==========================================================================================

    def __read_global_config(self):

        pax_config_path = self.global_config_dst
        log.info("Loading Config Path", {
            "path": pax_config_path
        })

        with open(pax_config_path, "r") as f:
            data = yaml.load(f)

        log.info("Local Config Loaded", {
            "path": pax_config_path,
            "data": data
        })

        return data



# Execute all actions to 'deploy' this repo.
from .pypax import Pax
from logkit import log
from .pax_action import PaxAction
import os
import yaml

# ==========================================================================================
# Import built-in actions.
# ==========================================================================================

from .pax_action_modify_helm_image import PaxActionModifyHelmImage

# ==========================================================================================
# Constants.
# ==========================================================================================

TAG_DEFAULT = "default"

# ==========================================================================================
# Methods.
# ==========================================================================================


def deploy(pax: Pax, tag: str="staging"):

    actions = load_actions(pax)
    steps = load_deployment(pax, tag)
    env_map = load_environment_config(pax, tag)

    for i, key in enumerate(steps):
        if key not in actions:
            log.warning("Action not found", {
                "step_i": i,
                "action_key": key
            })
            break
        
        action = actions[key]
        action.execute(env_map)


def load_environment_config(pax: Pax, tag: str="default"):

    # Load the default environment map first.
    pax_data = pax.load_pax_actions_data()
    deploy_data = pax_data["deploy"]
    env_map = deploy_data[TAG_DEFAULT]["config"]  # Load the default.

    # Update/override the environment map with values from the overarching directive.
    if tag in deploy_data and tag != TAG_DEFAULT:
        tag_env_map = deploy_data[tag]["config"]
        for k, v in tag_env_map.items():
            env_map[k] = v

    # Updated env map.
    log.info("Loaded ENV Config", env_map)
    return env_map


def load_actions(pax: Pax):
    """ Load all the valid actions that can be deployed. """

    # Start with the hard-coded actions.
    actions = __load_hard_coded_actions()

    # Load action definitions from the yaml.
    actions_map = pax.load_pax_actions_data()["actions"]
    if actions_map is None:
        log.warning("No actions found.", {"path": "actions.yml"})

    # Link each action.
    for k, v in actions_map.items():
        action = PaxAction(k, v)
        actions[k] = action
    
    # Return the final map.
    return actions


def __load_hard_coded_actions():
    """ Load hard-coded Python actions into the map. """
    actions = {
        "modify_helm_image": PaxActionModifyHelmImage("modify_helm_image")
    }
    return actions


def load_deployment(pax: Pax, deploy_tag: str):
    """ Load all the saved actions from this Pax instance. """

    # Get the deployment object from the PAX yaml.
    deploy_map = pax.load_pax_actions_data()["deploy"]
    steps = []

    if deploy_map is None:
        log.warning("No actions found.", {"path": pax.actions_data_dst})
    else:
        log.info("Deployment Object Loaded", deploy_map)
        steps = deploy_map[deploy_tag]["steps"]

    return steps



__version__ = "0.0.57"

from .pypax import Pax
from . import pax_deploy
from . import pax_global
from . import pax_initialize


def instance():
    return Pax.instance()


def configure():
    pax_global.configure(instance())


def initialize():
    pax_initialize.initialize(instance())


def deploy():
    pax_deploy.deploy(instance())



from .pax_action import PaxAction
import os
from logkit import log
import yaml
from typing import Union
import re


class PaxActionModifyHelmImage(PaxAction):
    
    HELM_VALUES_FILE_KEY = "HELM_VALUES_FILE"
    ECR_REPO_KEY = "ECR_REPO"
    DOCKER_IMAGE_KEY = "DOCKER_IMAGE"
    COMMIT_TAG_FILE = ".pax/commit_hash"

    def get_docker_tag(self):
        """ Extract the Docker commit tag from the local file system. """
        tag = None
        with open(self.COMMIT_TAG_FILE, "r") as f:
            tag = f.read()
        
        tag = tag.strip().replace("\n", "")
        log.info("Got Docker Tag", {"tag": tag})
        return tag

    def get_docker_image(self, env_map: dict):
        """ Extract the Docker repo/image from the ENV vars. """
        repo = env_map[self.ECR_REPO_KEY]
        image = env_map[self.DOCKER_IMAGE_KEY]
        full_repo_path = "{}/{}".format(repo, image)
        log.info("Got Docker Repo", {"repo": full_repo_path})
        return full_repo_path

    def execute(self, env_map: Union[dict, None]=None):
        
        values_file = env_map[self.HELM_VALUES_FILE_KEY]  # "./fake_helm/values.fake.yaml"  # env_map["HELM_VALUES_FILE"]
        log.info("Helm Values File", {
            "path": values_file
        })

        # Attempt to update values file.
        if not os.path.exists(values_file):
            log.error("Values File Not Found!", {
                "file": values_file
            })
            raise Exception("BOOM")

        replace_data = {
            "repository": self.get_docker_image(env_map),
            "tag": self.get_docker_tag()
        }

        # Attempt line-by-line update to keep the rest of the file stable.
        with open(values_file, "r") as f:
            lines = f.readlines()

        with open("./fake_helm/values.out.yaml", "w") as f:
            
            under_image = False
            image_modified = False

            for line in lines:

                line_written = False

                if not under_image and not image_modified:
                    is_image = re.search("^image:", line)
                    if is_image:
                        under_image = True
                else:
                    if under_image:
                        # We either find a matching line, or we break the pattern.
                        is_attribute = re.search("^ ", line) is not None
                        if is_attribute:
                            attribute_heading = line.split(":")[0]
                            attribute_name = attribute_heading.strip()
                            print("Got Attribute: ", attribute_name)
                            if attribute_name in replace_data:
                                replace_str = "{}: {}\n".format(attribute_heading, replace_data[attribute_name])
                                f.write(replace_str)
                                line_written = True
                        else:
                            under_image = False
                            image_modified = True

                if not line_written:
                    f.write(line)

        result = "Helm Chart Updated"

        self.log_results(
            self.name, 
            "<REDACTED>", 
            result)



from logkit import log
import os
from typing import Union

class PaxAction:

    """ A container for an action within a specific repo. """

    def __init__(self, name, commands=None):
        self.name = name
        self.commands = commands
    
    def execute(self, env_map: Union[dict, None]=None):

        # Join the commands into a full script.
        full_script = "\n".join(self.commands)
        if env_map is not None:
            full_script = self.append_with_environment(full_script, env_map)
        os.system(full_script)

        # Check if results were written to the buffer.
        line = None
        tmp_file = ".tmp_result.txt"
        if os.path.exists(tmp_file):
            with open(tmp_file, "r") as f:
                line = f.readline().strip().replace("\n", "")
            os.remove(tmp_file)
        
        # Print action execution summary.
        self.log_results(self.name, {i: v for i, v in enumerate(self.commands)}, line)

    def log_results(self, name, commands, result):
        log.info("Execution Action", {
            "name": name,
            "commands": commands,
            "result": result
        })

    def append_with_environment(self, script: str, env_map: dict):
        """ Append the environment variables to the script before running it. """
        env_arr = []
        for k, v in env_map.items():
            env_cmd = "{}={};".format(k, v)
            env_arr.append(env_cmd)
        
        env_str = " ".join(env_arr)
        final_command = "{} {}".format(env_str, script)
        return final_command



from .pypax import Pax
from logkit import log
import os
import yaml
import shutil


# ==========================================================================================
# Public Methods.
# ==========================================================================================

def initialize(pax: Pax):
    """ Initialize this path as a PyPAX source. """

    log.info("Initializing Repo as PAX package!")
    tag = __enter_verified_tag(pax)
    description = input("Enter package description: ")
    link = pax.get_remote_git_branch()

    node = {
        "tag": tag,
        "inbound": [],
        "version": "0.0.0",
        "description": description,
        "link": link
    }

    node_data = {
        "node": node
    }

    log.info("Node Created", node)

    pax.add_node_to_server(node)
    pax.save_local_object(pax.PAX_NODE_FILE, node_data)
    __clone_default_actions(pax)

    log.info("Added to server")


# ==========================================================================================
# Private Methods.
# ==========================================================================================


def __enter_verified_tag(pax: Pax):
    """ User enters a tag until they find one that is verified. """

    server_tags = pax.get_tags()

    while True:
        
        tag = input("Enter 3-letter tag for this package: ")
        if len(tag) != 3:
            log.warning("Tag must be 3 characters!", {
                "invalid_tag": tag
            })
            continue
        tag = tag[:3].upper()

        if tag in server_tags:
            log.warning("Tag already exists on the server! Please enter a different tag.", {
                "invalid_tag": tag
            })
            continue
        break

    return tag


def __clone_default_actions(pax: Pax):
    """ PyPAX comes bundled with a default pax_actions.yml. Clone this into the local repo. """
    dir_path = os.path.dirname(os.path.realpath(__file__))
    print(dir_path)
    actions_file = os.path.join(dir_path, "pax_actions.yml")
    shutil.copy(actions_file, pax.actions_data_dst)
    


from .pypax import Pax
from logkit import log
import os
import yaml


# ==========================================================================================
# Public Methods.
# ==========================================================================================


def configure(pax: Pax):
    log.info("Configuring PyPax Singleton!")

    # Create a configuration object.
    config_object = {}

    # Populate it with each of the configs.
    __configure_api_endpoint(pax, config_object)
    __configure_github(config_object)

    # Write the configuration object to disk.
    __write_config(pax, config_object)

    log.info("Pax Configuration Object", config_object)
    log.info("Pax Configuration Results", {
        "success": True,
        "destination": pax.global_config_dst
    })


# ==========================================================================================
# Private Methods.
# ==========================================================================================


def __configure_api_endpoint(pax: Pax, config_object):

    success = False
    while not success:

        # Ask user for the end-point.
        api_endpoint = input("Enter PyPAX Server Endpoint: ")
        log.info("API Endpoint", {"uri": api_endpoint})

        # Test endpoint.
        pax.api_endpoint = api_endpoint
        try:
            success = pax.test_connection()
        except Exception as e:
            log.warning("Unable to connect to endpoint", {
                "endpoint": api_endpoint,
                "error": str(e)
            })

        if success:
            # Add the API enpoint to the configuration object.
            config_object["api_endpoint"] = api_endpoint
        else:
            log.warning("This API Endpoint failed to connect.")


def __configure_github(config_object):
    """ Enter a valid GitHub username and token. """
    github_user = input("Enter GitHub User: ")
    github_pass = input("Enter GitHub Pass: ")
    config_object["github_user"] = github_user
    config_object["github_pass"] = github_pass


def __write_config(pax: Pax, config_object):
    """ Write the configuration object to disk. """
    pax_config_path = pax.global_config_dst
    with open(pax_config_path, "w") as f:
        yaml.dump(config_object, f)
